/* 
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Boehm, February 7, 1996 4:32 pm PST */
 
#include <stdio.h>
#include "gc_priv_new.h"

extern ptr_t GC_clear_stack();	/* in misc.c, behaves like identity */
void GC_extend_size_map();	/* in misc.c. */

/* Allocate reclaim list for kind:	*/
/* Return TRUE on success		*/
bool GC_alloc_reclaim_list(kind)
register struct obj_kind * kind;
{
    struct hblk ** result = (struct hblk **)
    		GC_scratch_alloc((MAXOBJSZ+1) * sizeof(struct hblk *));
    if (result == 0) return(FALSE);
    BZERO(result, (MAXOBJSZ+1)*sizeof(struct hblk *));
    kind -> ok_reclaim_list = result;
    return(TRUE);
}

/* allocate lb bytes for an object of kind.	*/
/* Should not be used to directly to allocate	*/
/* objects such as STUBBORN objects that	*/
/* require special handling on allocation.	*/
/* First a version that assumes we already	*/
/* hold lock:					*/
ptr_t GC_generic_malloc_inner(lb, k)
register word lb;
register int k;
{
register word lw;
register ptr_t op;
register ptr_t *opp;

    if( SMALL_OBJ(lb) ) {
        register struct obj_kind * kind = GC_obj_kinds + k;
#       ifdef MERGE_SIZES
	  lw = GC_size_map[lb];
#	else
	  lw = ALIGNED_WORDS(lb);
	  if (lw == 0) lw = 1;
#       endif
	opp = &(kind -> ok_freelist[lw]);
        if( (op = *opp) == 0 ) {
#	    ifdef MERGE_SIZES
	      if (GC_size_map[lb] == 0) {
	        if (!GC_is_initialized)  GC_init_inner();
	        if (GC_size_map[lb] == 0) GC_extend_size_map(lb);
	        return(GC_generic_malloc_inner(lb, k));
	      }
#	    else
	      if (!GC_is_initialized) {
	        GC_init_inner();
	        return(GC_generic_malloc_inner(lb, k));
	      }
#	    endif
	    if (kind -> ok_reclaim_list == 0) {
	    	if (!GC_alloc_reclaim_list(kind)) goto out;
	    }
	    op = GC_allocobj(lw, k);
	    if (op == 0) goto out;
        }
        /* Here everything is in a consistent state.	*/
        /* We assume the following assignment is	*/
        /* atomic.  If we get aborted			*/
        /* after the assignment, we lose an object,	*/
        /* but that's benign.				*/
        /* Volatile declarations may need to be added	*/
        /* to prevent the compiler from breaking things.*/
        *opp = obj_link(op);
        obj_link(op) = 0;
    } else {
	register struct hblk * h;
	register word n_blocks = divHBLKSZ(ADD_SLOP(lb)
					   + HDR_BYTES + HBLKSIZE-1);
	
	if (!GC_is_initialized) GC_init_inner();
	/* Do our share of marking work */
          if(GC_incremental && !GC_dont_gc)
		GC_collect_a_little_inner((int)n_blocks);
	lw = ROUNDED_UP_WORDS(lb);
/******* REMOVED
	while ((h = GC_allochblk(lw, k, 0)) == 0
		&& GC_collect_or_expand(n_blocks, FALSE));
*********/
/******** NEW CODE */
	if ((h = GC_allochblk(lw, k, 0)) == 0) {
		GC_collect_or_expand(n_blocks, FALSE, TRUE);
		if ((h = GC_allochblk(lw, k, 0)) == 0) {
			GC_collect_or_expand(n_blocks, FALSE, FALSE);
			if ((h = GC_allochblk(lw, k, 0)) == 0) {
				printf("Out of memory, giving up\n");
				exit(-1);
			}
		}
	}
/*********/
	if (h == 0) {
	    op = 0;
	} else {
	    op = (ptr_t) (h -> hb_body);
	    GC_words_wasted += BYTES_TO_WORDS(n_blocks * HBLKSIZE) - lw;
	}
    }
    GC_words_allocd += lw;
    
out:
    return((ptr_t)op);
}

/* Allocate a composite object of size n bytes.  The caller guarantees	*/
/* that pointers past the first page are not relevant.  Caller holds	*/
/* allocation lock.							*/
ptr_t GC_generic_malloc_inner_ignore_off_page(lb, k)
register size_t lb;
register int k;
{
    register struct hblk * h;
    register word n_blocks;
    register word lw;
    register ptr_t op;

    if (lb <= HBLKSIZE)
        return(GC_generic_malloc_inner((word)lb, k));
    n_blocks = divHBLKSZ(ADD_SLOP(lb) + HDR_BYTES + HBLKSIZE-1);
    if (!GC_is_initialized) GC_init_inner();
    /* Do our share of marking work */
    if(GC_incremental && !GC_dont_gc)
	GC_collect_a_little_inner((int)n_blocks);
    lw = ROUNDED_UP_WORDS(lb);
/******** REMOVED
    while ((h = GC_allochblk(lw, k, IGNORE_OFF_PAGE)) == 0
	   && GC_collect_or_expand(n_blocks, TRUE));
*********/
/******** NEW CODE */
	if ((h = GC_allochblk(lw, k, IGNORE_OFF_PAGE)) == 0) {
		GC_collect_or_expand(n_blocks, FALSE, TRUE);
		if ((h = GC_allochblk(lw, k, IGNORE_OFF_PAGE)) == 0) {
			GC_collect_or_expand(n_blocks, FALSE, FALSE);
			if ((h = GC_allochblk(lw, k, IGNORE_OFF_PAGE)) == 0) {
				printf("Out of memory, giving up\n");
				exit(-1);
			}
		}
	}
/*********/
    if (h == 0) {
	op = 0;
    } else {
	op = (ptr_t) (h -> hb_body);
	GC_words_wasted += BYTES_TO_WORDS(n_blocks * HBLKSIZE) - lw;
    }
    GC_words_allocd += lw;
    return((ptr_t)op);
}

ptr_t GC_generic_malloc_ignore_off_page(lb, k)
register size_t lb;
register int k;
{
    register ptr_t result;
    DCL_LOCK_STATE;
    
    GC_invoke_finalizers();
    DISABLE_SIGNALS();
    LOCK();
    result = GC_generic_malloc_inner_ignore_off_page(lb,k);
    UNLOCK();
    ENABLE_SIGNALS();
    return(result);
}

# if defined(__STDC__) || defined(__cplusplus)
  void * GC_malloc_ignore_off_page(size_t lb)
# else
  char * GC_malloc_ignore_off_page(lb)
  register size_t lb;
# endif
{
    return((GC_PTR)GC_generic_malloc_ignore_off_page(lb, NORMAL));
}

# if defined(__STDC__) || defined(__cplusplus)
  void * GC_malloc_atomic_ignore_off_page(size_t lb)
# else
  char * GC_malloc_atomic_ignore_off_page(lb)
  register size_t lb;
# endif
{
    return((GC_PTR)GC_generic_malloc_ignore_off_page(lb, PTRFREE));
}

ptr_t GC_generic_malloc(lb, k)
register word lb;
register int k;
{
    ptr_t result;
    DCL_LOCK_STATE;

    GC_invoke_finalizers();
    DISABLE_SIGNALS();
    LOCK();
    result = GC_generic_malloc_inner(lb, k);
    UNLOCK();
    ENABLE_SIGNALS();
    return(result);
}   


/* Analogous to the above, but assumes a small object size, and 	*/
/* bypasses MERGE_SIZES mechanism.  Used by gc_inline.h.		*/
ptr_t GC_generic_malloc_words_small(lw, k)
register word lw;
register int k;
{
register ptr_t op;
register ptr_t *opp;
register struct obj_kind * kind = GC_obj_kinds + k;
DCL_LOCK_STATE;

    GC_invoke_finalizers();
    DISABLE_SIGNALS();
    LOCK();
    opp = &(kind -> ok_freelist[lw]);
    if( (op = *opp) == 0 ) {
        if (!GC_is_initialized) {
            GC_init_inner();
        }
	if (kind -> ok_reclaim_list == 0) {
	    if (!GC_alloc_reclaim_list(kind)) goto out;
	}
	op = GC_clear_stack(GC_allocobj(lw, k));
	if (op == 0) goto out;
    }
    *opp = obj_link(op);
    obj_link(op) = 0;
    GC_words_allocd += lw;
    
out:
    UNLOCK();
    ENABLE_SIGNALS();
    return((ptr_t)op);
}

#if defined(THREADS) && !defined(SRC_M3)
/* Return a list of 1 or more objects of the indicated size, linked	*/
/* through the first word in the object.  This has the advantage that	*/
/* it acquires the allocation lock only once, and may greatly reduce	*/
/* time wasted contending for the allocation lock.  Typical usage would */
/* be in a thread that requires many items of the same size.  It would	*/
/* keep its own free list in thread-local storage, and call		*/
/* GC_malloc_many or friends to replenish it.  (We do not round up	*/
/* object sizes, since a call indicates the intention to consume many	*/
/* objects of exactly this size.)					*/
/* Note that the client should usually clear the link field.		*/
ptr_t GC_generic_malloc_many(lb, k)
register word lb;
register int k;
{
ptr_t op;
register ptr_t p;
ptr_t *opp;
word lw;
register word my_words_allocd;
DCL_LOCK_STATE;

    if (!SMALL_OBJ(lb)) {
        op = GC_generic_malloc(lb, k);
        obj_link(op) = 0;
        return(op);
    }
    lw = ALIGNED_WORDS(lb);
    GC_invoke_finalizers();
    DISABLE_SIGNALS();
    LOCK();
    opp = &(GC_obj_kinds[k].ok_freelist[lw]);
    if( (op = *opp) == 0 ) {
        if (!GC_is_initialized) {
            GC_init_inner();
        }
	op = GC_clear_stack(GC_allocobj(lw, k));
	if (op == 0) goto out;
    }
    *opp = 0;
    my_words_allocd = 0;
    for (p = op; p != 0; p = obj_link(p)) {
        my_words_allocd += lw;
        if (my_words_allocd >= BODY_SZ) {
            *opp = obj_link(p);
            obj_link(p) = 0;
            break;
        }
    }
    GC_words_allocd += my_words_allocd;
    
out:
    UNLOCK();
    ENABLE_SIGNALS();
    return(op);

}

void * GC_malloc_many(size_t lb)
{
    return(GC_generic_malloc_many(lb, NORMAL));
}

/* Note that the "atomic" version of this would be unsafe, since the	*/
/* links would not be seen by the collector.				*/
# endif

#define GENERAL_MALLOC(lb,k) \
    (GC_PTR)GC_clear_stack(GC_generic_malloc((word)lb, k))
/* We make the GC_clear_stack_call a tail call, hoping to get more of	*/
/* the stack.								*/

/* Allocate lb bytes of atomic (pointerfree) data */

/************ CHANGE *******************/
#ifdef PORTABLE_CLU
# ifdef __STDC__
    GC_PTR GC_malloc_atomic(size_t lb)
# else
    GC_PTR GC_malloc_atomic(lb)
    size_t lb;
# endif
{
CLUREF ans;
errcode err;

	clu_alloc_atomic(lb, &ans);
	return((struct obj *)ans.ref);
	}
void clu_alloc_atomic(lb, ans)
int lb;
char **ans;
#else
/************ END CHANGE ***************/
# ifdef __STDC__
    GC_PTR GC_malloc_atomic(size_t lb)
# else
    GC_PTR GC_malloc_atomic(lb)
    size_t lb;
# endif
/************ CHANGE *******************/
# endif
/************ END CHANGE ***************/
{
register ptr_t op;
register ptr_t * opp;
register word lw;
DCL_LOCK_STATE;

    if( SMALL_OBJ(lb) ) {
#       ifdef MERGE_SIZES
	  lw = GC_size_map[lb];
#	else
	  lw = ALIGNED_WORDS(lb);
#       endif
	opp = &(GC_aobjfreelist[lw]);
	FASTLOCK();
        if( !FASTLOCK_SUCCEEDED() || (op = *opp) == 0 ) {
            FASTUNLOCK();
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = GENERAL_MALLOC((word)lb, PTRFREE);
	    bzero((char *)*ans, lb);
	    return;
#else
/*********** END CHANGE ***************/
            return(GENERAL_MALLOC((word)lb, PTRFREE));
/************ CHANGE *******************/
#endif
/*********** END CHANGE ***************/
        }
        /* See above comment on signals.	*/
        *opp = obj_link(op);
        GC_words_allocd += lw;
        FASTUNLOCK();
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = op;
	    bzero((char *)op, lb);
	    return;
#else
/*********** END CHANGE ***************/
        return((GC_PTR) op);
/************ CHANGE *******************/
#endif
/*********** END CHANGE ***************/
   } else {
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = GENERAL_MALLOC((word)lb, PTRFREE);
	    bzero((char *)*ans, lb);
	    return;
#else
/*********** END CHANGE ***************/
       return(GENERAL_MALLOC((word)lb, PTRFREE));
/************ CHANGE *******************/
#endif
/*********** END CHANGE ***************/
   }
}

/* Allocate lb bytes of composite (pointerful) data */
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
# ifdef __STDC__
    GC_PTR GC_malloc(size_t lb)
# else
    GC_PTR GC_malloc(lb)
    size_t lb;
# endif
{
CLUREF ans;
errcode err;

	clu_alloc(lb, &ans);
	return((struct obj *)ans.ref);
	}
void clu_alloc(lb, ans)
int lb;
char **ans;
#else
/************ END CHANGE ***************/
# ifdef __STDC__
    GC_PTR GC_malloc(size_t lb)
# else
    GC_PTR GC_malloc(lb)
    size_t lb;
# endif
/************ CHANGE *******************/
# endif
/************ END CHANGE ***************/
{
register ptr_t op;
register ptr_t *opp;
register word lw;
DCL_LOCK_STATE;

    if( SMALL_OBJ(lb) ) {
#       ifdef MERGE_SIZES
	  lw = GC_size_map[lb];
#	else
	  lw = ALIGNED_WORDS(lb);
#       endif
	opp = &(GC_objfreelist[lw]);
	FASTLOCK();
        if( !FASTLOCK_SUCCEEDED() || (op = *opp) == 0 ) {
            FASTUNLOCK();
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = GENERAL_MALLOC((word)lb, NORMAL);
	    bzero((char *)*ans, lb);
	    return;
#else
/*********** END CHANGE ***************/
            return(GENERAL_MALLOC((word)lb, NORMAL));
/************ CHANGE *******************/
# endif
/************ END CHANGE ***************/
        }
        /* See above comment on signals.	*/
        *opp = obj_link(op);
        obj_link(op) = 0;
        GC_words_allocd += lw;
        FASTUNLOCK();
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = op;
	    bzero((char *)op, lb);
	    return;
#else
/*********** END CHANGE ***************/
        return((GC_PTR) op);
/************ CHANGE *******************/
# endif
/************ END CHANGE ***************/
   } else {
/************ CHANGE *******************/
#ifdef PORTABLE_CLU
	    *ans = GENERAL_MALLOC((word)lb, NORMAL);
	    bzero((char *)*ans, lb);
	    return;
#else
/*********** END CHANGE ***************/
       return(GENERAL_MALLOC((word)lb, NORMAL));
/************ CHANGE *******************/
# endif
/************ END CHANGE ***************/
   }
}

# ifdef REDIRECT_MALLOC
# ifdef __STDC__
    GC_PTR malloc(size_t lb)
# else
    GC_PTR malloc(lb)
    size_t lb;
# endif
  {
    /* It might help to manually inline the GC_malloc call here.	*/
    /* But any decent compiler should reduce the extra procedure call	*/
    /* to at most a jump instruction in this case.			*/
    return(REDIRECT_MALLOC(lb));
  }

# ifdef __STDC__
    GC_PTR calloc(size_t n, size_t lb)
# else
    GC_PTR calloc(n, lb)
    size_t n, lb;
# endif
  {
    return(REDIRECT_MALLOC(n*lb));
  }
# endif /* REDIRECT_MALLOC */

/* Allocate lb bytes of pointerful, traced, but not collectable data */
# ifdef __STDC__
    GC_PTR GC_malloc_uncollectable(size_t lb)
# else
    GC_PTR GC_malloc_uncollectable(lb)
    size_t lb;
# endif
{
register ptr_t op;
register ptr_t *opp;
register word lw;
DCL_LOCK_STATE;

    if( SMALL_OBJ(lb) ) {
#       ifdef MERGE_SIZES
#	  ifdef ADD_BYTE_AT_END
	    if (lb != 0) lb--;
	    	  /* We don't need the extra byte, since this won't be	*/
	    	  /* collected anyway.					*/
#	  endif
	  lw = GC_size_map[lb];
#	else
	  lw = ALIGNED_WORDS(lb);
#       endif
	opp = &(GC_uobjfreelist[lw]);
	FASTLOCK();
        if( FASTLOCK_SUCCEEDED() && (op = *opp) != 0 ) {
            /* See above comment on signals.	*/
            *opp = obj_link(op);
            obj_link(op) = 0;
            GC_words_allocd += lw;
            GC_set_mark_bit(op);
            GC_non_gc_bytes += WORDS_TO_BYTES(lw);
            FASTUNLOCK();
            return((GC_PTR) op);
        }
        FASTUNLOCK();
        op = (ptr_t)GC_generic_malloc((word)lb, UNCOLLECTABLE);
    } else {
	op = (ptr_t)GC_generic_malloc((word)lb, UNCOLLECTABLE);
    }
    /* We don't need the lock here, since we have an undisguised 	*/
    /* pointer.  We do need to hold the lock while we adjust		*/
    /* mark bits.							*/
    {
	register struct hblk * h;
	
	h = HBLKPTR(op);
	lw = HDR(h) -> hb_sz;
	
	DISABLE_SIGNALS();
	LOCK();
	GC_set_mark_bit(op);
	GC_non_gc_bytes += WORDS_TO_BYTES(lw);
	UNLOCK();
	ENABLE_SIGNALS();
	return((GC_PTR) op);
    }
}

GC_PTR GC_generic_or_special_malloc(lb,knd)
word lb;
int knd;
{
    switch(knd) {
#     ifdef STUBBORN_ALLOC
	case STUBBORN:
	    return(GC_malloc_stubborn((size_t)lb));
#     endif
	case PTRFREE:
	    return(GC_malloc_atomic((size_t)lb));
	case NORMAL:
	    return(GC_malloc((size_t)lb));
	case UNCOLLECTABLE:
	    return(GC_malloc_uncollectable((size_t)lb));
	default:
	    return(GC_generic_malloc(lb,knd));
    }
}


/* Change the size of the block pointed to by p to contain at least   */
/* lb bytes.  The object may be (and quite likely will be) moved.     */
/* The kind (e.g. atomic) is the same as that of the old.	      */
/* Shrinking of large blocks is not implemented well.                 */
# ifdef __STDC__
    GC_PTR GC_realloc(GC_PTR p, size_t lb)
# else
    GC_PTR GC_realloc(p,lb)
    GC_PTR p;
    size_t lb;
# endif
{
register struct hblk * h;
register hdr * hhdr;
register word sz;	 /* Current size in bytes	*/
register word orig_sz;	 /* Original sz in bytes	*/
int obj_kind;

    if (p == 0) return(GC_malloc(lb));	/* Required by ANSI */
    h = HBLKPTR(p);
    hhdr = HDR(h);
    sz = hhdr -> hb_sz;
    obj_kind = hhdr -> hb_obj_kind;
    sz = WORDS_TO_BYTES(sz);
    orig_sz = sz;

    if (sz > WORDS_TO_BYTES(MAXOBJSZ)) {
	/* Round it up to the next whole heap block */
	  register word descr;
	  
	  sz = (sz+HDR_BYTES+HBLKSIZE-1)
		& (~HBLKMASK);
	  sz -= HDR_BYTES;
	  hhdr -> hb_sz = BYTES_TO_WORDS(sz);
	  descr = GC_obj_kinds[obj_kind].ok_descriptor;
          if (GC_obj_kinds[obj_kind].ok_relocate_descr) descr += sz;
          hhdr -> hb_descr = descr;
	  if (obj_kind == UNCOLLECTABLE) GC_non_gc_bytes += (sz - orig_sz);
	  /* Extra area is already cleared by allochblk. */
    }
    if (ADD_SLOP(lb) <= sz) {
	if (lb >= (sz >> 1)) {
#	    ifdef STUBBORN_ALLOC
	        if (obj_kind == STUBBORN) GC_change_stubborn(p);
#	    endif
	    if (orig_sz > lb) {
	      /* Clear unneeded part of object to avoid bogus pointer */
	      /* tracing.					      */
	      /* Safe for stubborn objects.			      */
	        BZERO(((ptr_t)p) + lb, orig_sz - lb);
	    }
	    return(p);
	} else {
	    /* shrink */
	      GC_PTR result =
	      		GC_generic_or_special_malloc((word)lb, obj_kind);

	      if (result == 0) return(0);
	          /* Could also return original object.  But this 	*/
	          /* gives the client warning of imminent disaster.	*/
	      BCOPY(p, result, lb);
	      GC_free(p);
	      return(result);
	}
    } else {
	/* grow */
	  GC_PTR result =
	  	GC_generic_or_special_malloc((word)lb, obj_kind);

	  if (result == 0) return(0);
	  BCOPY(p, result, sz);
	  GC_free(p);
	  return(result);
    }
}

# ifdef REDIRECT_MALLOC
# ifdef __STDC__
    GC_PTR realloc(GC_PTR p, size_t lb)
# else
    GC_PTR realloc(p,lb)
    GC_PTR p;
    size_t lb;
# endif
  {
    return(GC_realloc(p, lb));
  }
# endif /* REDIRECT_MALLOC */

/* Explicitly deallocate an object p.				*/
# ifdef __STDC__
    void GC_free(GC_PTR p)
# else
    void GC_free(p)
    GC_PTR p;
# endif
{
    register struct hblk *h;
    register hdr *hhdr;
    register signed_word sz;
    register ptr_t * flh;
    register int knd;
    register struct obj_kind * ok;
    DCL_LOCK_STATE;

    if (p == 0) return;
    	/* Required by ANSI.  It's not my fault ...	*/
    h = HBLKPTR(p);
    hhdr = HDR(h);
    knd = hhdr -> hb_obj_kind;
    sz = hhdr -> hb_sz;
    ok = &GC_obj_kinds[knd];
    if (sz <= MAXOBJSZ) {
#	ifdef THREADS
	    DISABLE_SIGNALS();
	    LOCK();
#	endif
	GC_mem_freed += sz;
	/* A signal here can make GC_mem_freed and GC_non_gc_bytes	*/
	/* inconsistent.  We claim this is benign.			*/
	if (knd == UNCOLLECTABLE) GC_non_gc_bytes -= WORDS_TO_BYTES(sz);
	if (ok -> ok_init) {
	    BZERO((word *)p + 1, WORDS_TO_BYTES(sz-1));
	}
	flh = &(ok -> ok_freelist[sz]);
	obj_link(p) = *flh;
	*flh = (ptr_t)p;
#	ifdef THREADS
	    UNLOCK();
	    ENABLE_SIGNALS();
#	endif
    } else {
    	DISABLE_SIGNALS();
        LOCK();
        GC_mem_freed += sz;
	if (knd == UNCOLLECTABLE) GC_non_gc_bytes -= WORDS_TO_BYTES(sz);
        GC_freehblk(h);
        UNLOCK();
        ENABLE_SIGNALS();
    }
}

# ifdef REDIRECT_MALLOC
#   ifdef __STDC__
      void free(GC_PTR p)
#   else
      void free(p)
      GC_PTR p;
#   endif
  {
      GC_free(p);
  }
# endif  /* REDIRECT_MALLOC */

/*
 * Paramters used to control growth of the heap and frequency of
 * garbage collection.
 *
 * A block of HBLKSIZE bytes is considered "active" is some (possibly
 * garbage) objects reside on it.
 *
 * blks_hard_limit is an upper bound on the number of active blocks.
 * When allocation would exceed blks_soft_limit, a gc is performed.
 *
 * Required: blks_soft_limit <= blks_hard_limit
 * Sensible: blks_active <= blks_soft_limit
 */

# define MAX_HEAP_DEFAULT MAXHBLKS /* default upper bound on heap, in blocks */
# define FIRST_GC 128           /* default threshold for first GC, in blocks */

                                /* How to updates GC threshold a' la VAX CLU */
# define add_DEFAULT 0Xe0
# define mask_DEFAULT 0X7f

int blks_active = 0;
/* int blks_hard_limit = MAX_HEAP_DEFAULT; */
int blks_hard_limit = 0;
int bytes_hard_limit = 0;
int blks_min_gc = FIRST_GC;
int blks_soft_limit = FIRST_GC;
int bytes_soft_limit = mulHBLKSZ(FIRST_GC);

static int addHBLK = add_DEFAULT;
static int maskHBLK = mask_DEFAULT;

/* adjust add & mask */
/* 1/24/94 mtv: {_}gcOPcontrol renamed to _gc_control */
errcode _gc_control(add, mask, elist)
CLUREF add;
CLUREF mask;
errlist elist;
{
  if (add.num < 0) signal(ERR_illegal);
  if (mask.num > add.num) signal(ERR_illegal);
  if (mask.num < HBLKSIZE) signal(ERR_illegal);
  addHBLK = divHBLKSZ(add.num);
  maskHBLK = divHBLKSZ(mask.num);
  signal(ERR_ok);
}

/* recompute next_gc */
void recompute_ngc()
{
int arm1 = 0;
int arm2 = 0;
int arm3 = 0;

  blks_hard_limit = divHBLKSZ(GC_get_max_heap_size());
  bytes_hard_limit = mulHBLKSZ(blks_hard_limit);
  blks_active = divHBLKSZ(GC_heapsize);
  if (addHBLK == 0 && maskHBLK == 0)
    blks_soft_limit = ((blks_active+1+add_DEFAULT) & ~mask_DEFAULT) -1;
  else
    blks_soft_limit = ((blks_active+1+addHBLK) & ~maskHBLK) -1;

  if (blks_active == 0 || (float) (blks_soft_limit - blks_active) / (float) blks_active < 1.0 ) {
    arm1 = 1;
    blks_soft_limit = 2.0 * blks_active; }

  if (blks_soft_limit - blks_active < 1024) {
    arm2 = 1;
    blks_soft_limit = blks_active + 1024;}

  if (blks_soft_limit > blks_hard_limit) {
    arm3 = 1;
    blks_soft_limit = blks_hard_limit;}

  if (blks_soft_limit < blks_min_gc)
    blks_soft_limit = blks_min_gc;

  bytes_soft_limit = mulHBLKSZ(blks_soft_limit);

/*  printf("INUSE %d SOFT %d %d %d mult_arm %d low_pass %d high_pass %d\n",
		blks_active, 
                blks_soft_limit, bytes_soft_limit, blks_hard_limit,
		arm1, arm2, arm3); */
/* INUSE 168 SOFT 128 524288 0 mult_arm 0 low_pass 1 high_pass 1 */

/*
INUSE 16 SOFT 1040 4259840 2500 mult_arm 0 low_pass 1 high_pass 0
INUSE 16 SOFT 1040 4259840 2500 mult_arm 0 low_pass 1 high_pass 0
Switched to incremental mode
Emulating dirty bits with mprotect/signals
returning TRUE 4984832 4259840 2
INUSE 1531 SOFT 2500 10240000 2500 mult_arm 1 low_pass 0 high_pass 1
*/

}

errcode _set_min_gc(n)
CLUREF n;
{
        if (n.num > (blks_hard_limit * HBLKSIZE)) signal(ERR_toobig);
        if (n.num < 0) signal(ERR_negative_size);
        blks_min_gc = divHBLKSZ(n.num);
        signal(ERR_ok);
        }

int GC_should_collect()
{
   if (GC_heapsize > bytes_soft_limit) {
#ifdef  PRINTSTATS
	GC_printf0("returning 1\n");
#endif
	return 1;
	}
   if (blks_soft_limit == blks_hard_limit) {
#ifdef  PRINTSTATS
	GC_printf0("returning 2\n");
#endif
	return 2;
	}
#ifdef  PRINTSTATS
	GC_printf0("returning 0\n");
#endif
   return 0;
}
